#include "divers.h"
#include <stdio.h>

/* Affichage des tables de transition et mission */
void affiche_tables (sdg_ht *ht_emis, double **t_trans, int col, int nb_cat) {
  int i,j;
  int deb,fin,pas;
  char **mots;
  int nb_mots;
  double proba;

  pas = (int)nb_cat/col;
  if ((nb_cat % col) != 0)
    pas ++;
  fin = 0;
  do {
    fin += pas;
    deb = fin - pas;
    if (fin > nb_cat)
      fin = nb_cat;
    printf("de %i a %i\n",deb,fin);
    if (t_trans != NULL) {
      printf("Table transition (*100):\n");
      printf("  |");
      for (i=deb ; i<fin ; i++) {
        printf("%2i  |",i);
      }
      printf("\n");
      for (i=0 ; i<nb_cat ; i++) {
        printf("%2i|",i);
        for (j=deb ; j<fin ; j++) {
          proba = t_trans[i][j]*100;
          if (proba == 0)
            printf("   0|");
          else if (proba < 10)
            printf("%.2f|",proba);
          else if (proba < 100)
            printf("%.1f|",proba);
          else
            printf(" %.0f|",proba);
        }
        printf("\n");
      }
      printf("  |");
      for (i=deb ; i<fin ; i++) {
        printf("%2i  |",i);
      }
      printf("\n");
    }
    if (ht_emis != NULL) {
      nb_mots = 5;
      mots = malloc(nb_mots*sizeof(char*));
      mots[0] = "<s>";
      mots[1] = "the";
      mots[2] = "Fulton";
      mots[3] = "County";
      mots[4] = "</s>";
      printf("Table d'emission partielle (*100): ");
      for (i=0 ; i<nb_mots ; i++)
        printf("%s,",mots[i]);
      for (j=0 ; j<nb_mots ; j++) {
        printf("\n%c%c%c",mots[j][0],mots[j][1],mots[j][2]);
        for (i=deb ; i<fin ; i++) {
          proba = proba_mot_cat(ht_emis,mots[j],i)*100;
          if (proba == 0)
            printf("   0|");
          else if (proba < 10)
            printf("%.2f|",proba);
          else
            printf("%.1f|",proba);
        }
      }
      printf("\n");
    }
  } while (fin != nb_cat);
}


/* Calcule la proba forward d'une phrase */
double proba_alpha (sdg_ht *ht_emis, double **t_trans, char **phrase, int long_phrase, int nb_cat) {
  double **alpha,proba_totale;
  int i,j,t;

  /** Initialisation de alpha */
  alpha = malloc(long_phrase*sizeof(double*));
  for (t=0 ; t<long_phrase ; t++) {
    alpha[t] = calloc(nb_cat,sizeof(double));
  }
  /*
   * alpha[0][i] = pi(i) pour tout i=0..nb_cat-1
   * Si on ne fait pas d'hypothse sur la structure d'une phrase (solution choisie):
   *   pi(i) = 1/nb_cat pour tout i=0..nb_cat-1 (quiprobabilit)
   * Mais on aurait pu faire l'hypothse qu'une phrase ne peut commencer que par '<s>', alors:
   *   pi(i) = {0 pour tout i!=44 (ZTERM)
   *           {1 pour i=44
   */
  for (i=0 ; i<nb_cat ; i++) {
    alpha[0][i] = 1/((double)nb_cat);
  }
  /** Etapes rcursives */
  /*
   * alpha[t][j] = sigma {alpha[t-1][i].E(i,m[t-1]).T(i,j)}
   *             i=0..nb_cat-1
   * Pour:
   *  t=1..long_phrase-1
   *  j=0..nb_cat-1
   * avec:
   *  E(i,m[t-1]) = proba_mot_cat(ht_emis, phrase[t-1], i)
   *  T(i,j) = proba_cat_cat(t_trans,i,j)
   */
  //affiche_tables(ht_emis,t_trans,1,nb_cat);
  for (t=1 ; t<long_phrase ; t++) {
    for (j=0 ; j<nb_cat ; j++) {
      for (i=0 ; i<nb_cat ; i++) {
        alpha[t][j] += alpha[t-1][i]*proba_mot_cat(ht_emis,phrase[t-1],i)*proba_cat_cat(t_trans,i,j);
      }
      /*
      if (alpha[t][j] == 0)
        printf("alpha[%i][%i]=0|",t,j);
      else
        alpha[t][j] *= 100; // sinon alpha est trop petit
      */
    }
    //printf("\n");
  }
  /** Probabilit totale */
  /*
   * P(phrase) = sigma {alpha[T-1][i].E(i,m[T-1])}
   *           i=0..nb_cat-1
   * avec:
   *  T = long_phrase
   *  E(i,m[T-1]) = proba_mot_cat(ht_emis, phrase[T-1], i)
   */
  proba_totale = 0;
  for (i=0 ; i<nb_cat ; i++) {
    proba_totale += alpha[long_phrase-1][i]*proba_mot_cat(ht_emis, phrase[long_phrase-1], i);
  }
  /*
  if (proba_totale == 0) {
    printf("proba_totale=0 {\n");
    for (i=0 ; i<nb_cat ; i++) {
      printf("alpha[%i][%i]=%f ",long_phrase-1,i,alpha[long_phrase-1][i]);
      printf("E(%s,%i)=%f ",phrase[long_phrase-1],i,proba_mot_cat(ht_emis,phrase[long_phrase-1],i));
      printf("\n");
    }
    printf("}\n\n");
  }
  else 
    printf("proba_totale=%g\n\n",proba_totale);
  */
  /** Libration de la mmoire */
  for (t=0 ; t<long_phrase ; t++) {
    free(alpha[t]);
  }
  free(alpha);

  return proba_totale;
}

/* Calcule la proba backward d'une phrase */
double proba_beta (sdg_ht *ht_emis, double **t_trans, char **phrase, int long_phrase, int nb_cat) {
  double **beta,proba_totale;
  int i,j,t;

  /** Initialisation de beta */
  beta = malloc(long_phrase*sizeof(double*));
  for (t=0 ; t<long_phrase ; t++) {
    beta[t] = calloc(nb_cat,sizeof(double));
  }
  /*
   * beta[T-1][i] = E(i,m[T-1]) pour tout i=0..nb_cat-1
   * avec:
   *  T = long_phrase
   *  E(i,m[T-1]) = proba_mot_cat(ht_emis, phrase[T-1], i)
   */
  for (i=0 ; i<nb_cat ; i++) {
    beta[long_phrase-1][i] = proba_mot_cat(ht_emis, phrase[long_phrase-1], i);
  }
  /** Etapes rcursives */
  /*
   * beta[t][i] = sigma {beta[t+1][j].E(i,m[t]).T(i,j)}
   *             j=0..nb_cat-1
   * Pour:
   *  t=long_phrase-2..0
   *  i=0..nb_cat-1
   * avec:
   *  E(i,m[t]) = proba_mot_cat(ht_emis, phrase[t], i)
   *  T(i,j) = proba_cat_cat(t_trans,i,j)
   */
  for (t=long_phrase-2 ; t>=0 ; t--) {
    for (i=0 ; i<nb_cat ; i++) {
      for (j=0 ; j<nb_cat ; j++) {
        beta[t][i] += beta[t+1][j]*proba_mot_cat(ht_emis, phrase[t], i)*proba_cat_cat(t_trans,i,j);
      }
    }
  }
  /** Probabilit totale */
  /*
   * P(phrase) = sigma {beta[0][i].pi(i)}
   *           i=0..nb_cat-1
   * Si on ne fait pas d'hypothse sur la structure d'une phrase (solution choisie):
   *   pi(i) = 1/nb_cat pour tout i=0..nb_cat-1 (quiprobabilit)
   * Mais on aurait pu faire l'hypothse qu'une phrase ne peut commencer que par '<s>', alors:
   *   pi(i) = {0 pour tout i!=44 (ZTERM)
   *           {1 pour i=44
   */
  proba_totale = 0;
  for (i=0 ; i<nb_cat ; i++) {
    proba_totale += beta[0][i]/((double)nb_cat);
  }
  /** Libration de la mmoire */
  for (t=0 ; t<long_phrase ; t++) {
    free(beta[t]);
  }
  free(beta);

  return proba_totale;
}
