#include "bigram.h"

double** charge_bigram (char* nom_fichier_bigram,sdg_ht *ht_cat,int nb_cat) {
  FILE* fic;
  double ** tab_cat_cat;
  int i;
  int occurence, *code1, *code2;
  char *cat1, *cat2;

  /* Ouverture du fichier */
  if ((fic=fopen(nom_fichier_bigram,"r")) == NULL) {
    printf("Imposible d'ouvrir %s\n",nom_fichier_bigram);
    perror("fopen");
    exit (-1);
  }
  /* Creation du tableau des transitions */
  tab_cat_cat = malloc(nb_cat*sizeof(double*));
  for (i=0 ; i<nb_cat ; i++) {
    tab_cat_cat[i] = calloc(nb_cat,sizeof(double));
  }
  /* Parsage */
  cat1 = malloc(TAILLE_MAX_CAT*sizeof(char));
  cat2 = malloc(TAILLE_MAX_CAT*sizeof(char));
  while (fscanf(fic,"%i %s %s\n",&occurence,cat1,cat2) != EOF) {
    code1 = sdg_ht_recherche(ht_cat, cat1);
    code2 = sdg_ht_recherche(ht_cat, cat2);
    /* Mise a jour du couple (cat1,cat2) dans le tableau des bigrams
     * ie calcul de C(cat(i),cat(j)) */
    tab_cat_cat[*code1][*code2] += occurence;
  }
  /* fin */
  free(cat1);
  free(cat2);
  close(fic);
  return tab_cat_cat;
}

void calcule_proba_bigram(double **tab_bigram, int nb_cat) {
  int i,j;
  int compt;

  for (i=0 ; i<nb_cat ; i++) {
    compt = 0;
    for (j=0 ; j<nb_cat ; j++) {
      /* Calcule de C(cat(i)) */
      compt += tab_bigram[i][j];
    }
    for (j=0 ; j<nb_cat ; j++) {
      /* Calcul de
       *                    C(cat(i),cat(j))
       * p(cat(j)|cat(i)) = ----------------
       *                       C(cat(i))
       * Rem : p(cat(j)|cat(i))=0 si (cat(j),cat(i)) n'apparait pas
       * dans le corpus d'entrainement
       */
      tab_bigram[i][j] /= compt;
    }
  }
  return;
}

void calcule_proba_bigram_lisse (double **tab_bigram, int nb_cat) {
  int i,j;
  int compt;

  for (i=0 ; i<nb_cat ; i++) {
    compt = 0;
    for (j=0 ; j<nb_cat ; j++) {
      /* Calcule de C(cat(i)) */
      compt += tab_bigram[i][j];
    }
    for (j=0 ; j<nb_cat ; j++) {
      /* Calcule de
       *                    C(cat(i),cat(j))+1
       * p(cat(j)|cat(i)) = ------------------
       *                       C(cat(i))+V
       * avec V=nb de catgories
       */
      tab_bigram[i][j] += 1;
      tab_bigram[i][j] /= (compt+nb_cat);
    }
  }
  return;
}

double proba_cat_cat (double **tab_bigram, int c1, int c2) {
  /* Retourne p(c2|c1) */
  return tab_bigram[c1][c2];
}

double* proba_all_cat (double **tab_bigram, int cat, int nb_cat) {
  double *tab;
  int i;

  tab = malloc(sizeof(double)*nb_cat);
  for (i=0 ; i <nb_cat ; i++) {
    tab[i] = tab_bigram[i][cat];
  }
  return tab;
}

