#include "dico.h"

sdg_ht* charge_dico (char* nom_fichier_dico, sdg_ht* ht_cat, int nb_cat) {
  sdg_ht *ht_dico;
  char *mot;
  char *cat;
  FILE* fic;
  int occurence;
  int *p_code;

  /* Ouverture du fichier */
  if ((fic=fopen(nom_fichier_dico,"r")) == NULL) {
    printf("Imposible d'ouvrir %s\n",nom_fichier_dico);
    perror("fopen");
    exit (-1);
  }
  /* Creation de la table de hachage du dico */
  ht_dico = sdg_ht_cree(TAILLE_DICO_HT);
  /* Parsage */
  mot = malloc(TAILLE_MAX_MOT*sizeof(char));
  cat = malloc(TAILLE_MAX_CAT*sizeof(char));
  while (fscanf(fic,"%i %s %s\n",&occurence,mot,cat) != EOF) {
    mot = realloc(mot,strlen(mot)+1);
    cat = realloc(cat,strlen(cat)+1);
    p_code = sdg_ht_recherche(ht_cat, cat);
    /* Insertion du mot ds la ht du dico */
    ajoute_entree_dico(ht_dico,mot,*p_code,occurence,nb_cat);
    mot = malloc(TAILLE_MAX_MOT*sizeof(char));
    cat = malloc(TAILLE_MAX_CAT*sizeof(char));
  }

  free(mot);
  free(cat);
  fclose(fic);
  return ht_dico;
}

void ajoute_entree_dico (sdg_ht* ht_dico, char* mot, int cat, int cpt, int nb_cat) {
  double *tab_occ;
  int i;

  /* Verifie si le mot se trouve deja ds la ht */
  if ((tab_occ=(double*)sdg_ht_recherche(ht_dico,mot)) == NULL) {
    /* Le mot ne se trouve pas ds la ht */
    tab_occ = calloc(nb_cat,sizeof(double));
    sdg_ht_insere(ht_dico,mot,tab_occ);
  }
  tab_occ[cat] += (double)cpt;
  return;
}

void calcule_proba_dico (sdg_ht* ht_dico,int nb_cat) {
  double *cpt_cat;
  int i,j;
  sdg_lcn *pm;
  
  cpt_cat = calloc(nb_cat,sizeof(double)); /* initialise tous les 
                                              compteurs  0 */
  /* Parcours de la ht (donc de tous les mots) 
     pour calculer C(c) pour toutes les cat */
  for(i=0; i<ht_dico->taille; i++) {
    for(pm=ht_dico->tab[i] ; pm ; pm=pm->suiv) {
      for (j=0 ; j<nb_cat ; j++) {
        /* C(c) = sigma {C(m,c)}
         *     ts les mots m
         * C(m,c) tant le nombre de fois o m a
         *        t tiquet par c 
         */
        cpt_cat[j] += ((double*)pm->val)[j];
      }
    }
  }
  /* Affichage du nombre de mots tiquets par catgorie */
  printf("*Nombre de mots etiquetes par la categorie cat : C(cat)");
  for (i=0;i<nb_cat;i++) {
    if (i%6 == 0)
      printf("\n");
    printf("(%2i:%7.0f) ",i,cpt_cat[i]);
  }
  /* Parcours de la ht pour calculer p(m|c) */
  for(i=0; i<ht_dico->taille; i++) {
    for(pm=ht_dico->tab[i]; pm; pm=pm->suiv) {
      for (j=0 ; j<nb_cat ; j++) {
        if (cpt_cat[j] != 0) {
          /* p(m|c)=C(m,c)/C(c) */
          ((double*)pm->val)[j] /= cpt_cat[j];
          /* Rem:
           * ((double*)pm->val)[c] contenait le nombre de fois
           * o m a t tiquet par c (C(m,c)) et contient
           * maintenant p(m|c)
           */
        }
        else {
          /* Rem:
           * Si C(c)=0 (cpt_cat[c]==0) ie aucun mot n'apparait dans 
           * cette catgorie cela signifie que p(m|c)=0 pour tous
           * les mots. Donc ((double*)pm->val)[c] (qui contenait le
           * le nombre de fois o m a t tiquet par c et qui doit
           * maintenant contenir p(m|c)) contient dj la bonne valeur
           * (ie 0). Il n'y a donc rien a faire.
           */
        }
      }
    }
  }

  free(cpt_cat);
  return;
}

double proba_mot_cat (sdg_ht* ht_dico, char* mot, int cat){
  double *prob_cat;

  if ((prob_cat=(double*)sdg_ht_recherche(ht_dico,mot)) == NULL) {
    // La proba d'mission d'un mot inconnu est 0
    return 0;
  }

  return prob_cat[cat];
}

double* proba_mot_all_cat (sdg_ht* ht_dico, char* mot){
  return (double*)sdg_ht_recherche(ht_dico,mot);
}
