#include "cat.h"
#include "dico.h"
#include "bigram.h"
#include "viterbi.h"
#include "divers.h"
#include "baum_welsh.h"
#include <stdlib.h>

int main (int argc, char **argv) {
  double prob;
  int *p_test;
  int i,compt,t;
  sdg_ht *cat_ht, *dico_ht;
  double **bigram, *pi;
  int nb_cat;
  FILE *fic_in,*fic_out, *fic_in2;
  char *mot, *mot2,**phrase;
  int taille;
  int *meilleur_chemin,*meilleur_chemin_lisse;
  int best;
  char *categorie;
  char **tab_cat;
  int fin=0;
  int palier,progression;
  long total;
  double **alpha, **beta, **gamma, ***ro;
  char **texte;
  int taille_o;
  double perf,old_perf;

  /** Creation de la table de hachage des categories (cat.c) */
  cat_ht = lire_fichier_codes("../data/brown.tagset",&nb_cat);
  /* Cration du tableau num_cat -> nom_cat */
  tab_cat = sdg_ht_allcat(cat_ht,nb_cat);
  /* test... */
  printf("----------------------------------------------------------\n");
  printf("Creation des categorie (cat.c)\n");
  p_test = malloc(sizeof(int));
  printf("*Toutes les categories\n");
  sdg_ht_affiche_clefs(cat_ht);
  printf("\n*Quelques numeros de categorie d'apres le nom de la categorie\n");
  p_test = (int*) sdg_ht_recherche(cat_ht,"CD");
  printf("CD=%i\n",*p_test);
  p_test = (int*) sdg_ht_recherche(cat_ht,"NN");
  printf("NN=%i\n",*p_test);
  p_test = (int*) sdg_ht_recherche(cat_ht,"ZTRM");
  printf("ZTRM=%i\n",*p_test);
  printf("*Les categories d'apres leur numero");
  for (i=0 ; i<nb_cat ; i++) {
    if (i%6 == 0)
      printf("\n");
    printf("(%2i:%7s) ",i,tab_cat[i]);
  }

  /** Creation de la table de hachage des mots (table d'mission) */
  dico_ht = charge_dico("../data/dico", cat_ht, nb_cat);
  /* test... */
  printf("\n----------------------------------------------------------\n");
  printf("Calcul de la table d'emission (dico.c)\n");
  // sdg_ht_affiche_clefs(dico_ht);
  printf("*Av calcul (calcule_proba_dico) : C('the',cat) (nb de 'the' par categorie)");
  for (i=0 ; i<nb_cat ; i++) {
    if (i%5 == 0)
      printf("\n");
    prob = proba_mot_cat(dico_ht,"the",i);
    if (prob == 0)
      printf("(cat %2i:     0) ",i, prob);
    else
      printf("(cat %2i:%6.0f) ",i, prob);
  }
  printf("\n");
  calcule_proba_dico(dico_ht,nb_cat);
  printf("\n*Ap calcul (calcule_proba_dico): p('the'|cat)=C('the',cat)/C(cat)");
  for (i=0 ; i<nb_cat ; i++) {
    if (i%5 == 0)
      printf("\n");
    prob = proba_mot_cat(dico_ht,"the",i);
    if (prob == 0)
      printf("(cat %2i:     0) ",i, prob);
    else
      printf("(cat %2i:%.4f) ",i, prob);
  }
  printf("\n");

  /** Creation de la table des transitions */
  bigram = charge_bigram("../data/brown.train.2g",cat_ht,nb_cat);
  /* Test */
  printf("----------------------------------------------------------\n");
  printf("Calcul de la table des transitions (bigram.c)\n");
  compt = 0;
  printf("*Av calcul proba (calcule_proba_bigram) : C(1,c) ");
  printf("(nb de fois qu'un mot de cat 1\n est suivi d'un mot de cat c)");
  for (i=0 ; i<nb_cat ; i++) {
    compt += bigram[1][i];
    if (i%7 == 0)
      printf("\n");
    printf("(1-%2i=%4.0f) ",i,bigram[1][i]);
  }
  printf("\n*Nb de fois qu'un mot etiquete par 1 est suivi d'un ");
  printf("autre mot : C(1)\n");
  printf("C(1)=%i\n",compt);
  calcule_proba_bigram(bigram,nb_cat); // Calcul de T(i,j)
  printf("*Ap calcul proba : p(c|1)=C(1,c)/C(1)");
  for (i=0 ; i<nb_cat ; i++) {
    if (i%6 == 0)
      printf("\n");
    if (bigram[1][i] == 0)
      printf("(1-%2i=     0) ",i);
    else
      printf("(1-%2i=%.4f) ",i,bigram[1][i]);
  }

  /** Amlioration du MMC */
  /* Initialisation de pi */
  pi = malloc(nb_cat*sizeof(double));
  for (i=0 ; i<nb_cat ; i++) {
    pi[i] = 1/(double)nb_cat;
  }
  /* Constitution de la suite d'osvervables */
  if ((fic_in2=fopen("../data/app_seul","r")) == NULL) {
    printf("Imposible d'ouvrir le fichier test\n");
    perror("fopen");
    exit (-1);
  }
  taille_o = 5018; // nb de mots ds app_seul
  texte = malloc(taille_o*sizeof(char*));
  for (t=0 ; t<taille_o ; t++) {
    mot2 = malloc(TAILLE_MAX_MOT*sizeof(char));
    if (fscanf(fic_in2,"%s\n",mot2) == EOF) {
      taille_o = t;
      break;
    }
    texte[t] = mot2;
  }
  /* Init de alpha, beta, gamma et ro */
  alpha = malloc(taille_o*sizeof(double*));
  for (t=0 ; t<taille_o ; t++) {
    alpha[t] = calloc(nb_cat,sizeof(double));
  }
  beta = malloc(taille_o*sizeof(double*));
  for (t=0 ; t<taille_o ; t++) {
    beta[t] = calloc(nb_cat,sizeof(double));
  }
  gamma = malloc(taille_o*sizeof(double*));
  for (t=0 ; t<taille_o ; t++) {
    gamma[t] = calloc(nb_cat,sizeof(double));
  }
  ro = malloc(taille_o*sizeof(double**));
  for (t=0 ; t<taille_o ; t++) {
    ro[t] = malloc(nb_cat*sizeof(double*));
    for (i=0 ; i<nb_cat ; i++) {
      ro[t][i] = calloc(nb_cat,sizeof(double));
    }
  }
  /* Calcul des perf du MMC av amlioration */
  calcule_alpha(alpha,dico_ht,bigram,pi,texte,taille_o,nb_cat);
  calcule_beta(beta,dico_ht,bigram,texte,taille_o,nb_cat);
  perf = 0;
  for (i=0 ; i<nb_cat ; i++) {
    perf += alpha[1][i] * beta[1][i];
  }
  /* Boucle d'amlioration */
  do {
    printf("amelioration\n");
    old_perf = perf;
    // Amlioration
    calcule_gamma(gamma,alpha,beta,taille_o,nb_cat);
    printf("gamma\n");
    calcule_ro(ro,dico_ht,bigram,alpha,beta,texte,taille_o,nb_cat);
    printf("ro\n");
    new_pi(pi,gamma,nb_cat);
    printf("new_pi\n");
    new_emis(dico_ht,gamma,texte,taille_o,nb_cat);
    printf("new_emis\n");
    new_trans(bigram,gamma,ro,texte,taille_o,nb_cat);
    printf("new_trans\n");
    // Calcul de alpha et beta
    calcule_alpha(alpha,dico_ht,bigram,pi,texte,taille_o,nb_cat);
    calcule_beta(beta,dico_ht,bigram,texte,taille_o,nb_cat);
    // Calcul de P(o) ap amlioration
    perf = 0;
    for (i=0 ; i<nb_cat ; i++) {
      perf += alpha[1][i] * beta[1][i];
    }
  } while (perf > old_perf); // Noramlement il faudraut revenir au MMC d'avant

  //printf("----------------------------------------------------------\n");
  //affiche_tables(dico_ht,bigram_lisse,2,nb_cat);
  //affiche_tables(dico_ht,bigram,2,nb_cat);

  /*** Etiquetage du fichier test */
  printf("\n----------------------------------------------------------\n");
  /** Initialisation ... */
  /* Init phrase */
  phrase = malloc(TAILLE_MAX_PHRASE*sizeof(char*));
  /* Ouverture du fichier  de test */
  if ((fic_in=fopen("../data/test","r")) == NULL) {
    printf("Imposible d'ouvrir le fichier test\n");
    perror("fopen");
    exit (-1);
  }
  /* Ouverture du fichier resultat de l'tiquetage */
  if ((fic_out=fopen("../data/result_etiquetage_plus","w")) == NULL) {
    printf("Imposible d'ouvrir le fichier resultat\n");
    perror("fopen");
    exit (-1);
  }
  /** Lecture du fichier test et ecriture du resultat phrase par phrase */
  /* Prparation de la barre de progression */
  printf("Etiquetage du fichier ../data/test (resultat ds ../data/test_etiquete)\n");
  printf("Calculs de proba de phrase (resultat ds ../data/test_proba_phrases)");
  printf("\n|0%c                                          100%c|\n",'%','%');
  printf("|");
  fflush(stdout);
  palier = 0;
  progression = 0;
  fseek(fic_in,0,SEEK_END);
  total = ftell(fic_in);
  fseek(fic_in,0,SEEK_SET);
  /* Boucle d'tiquetage */
  while (1) {
    /* Lecture d'une phrase */
    taille = 0;
    do {
      mot = malloc(TAILLE_MAX_MOT*sizeof(char));
      if (fscanf(fic_in,"%s\n",mot) == EOF) {
        // La fin du fichier  tiqueter est atteinte -> fin
        fclose(fic_in);
        fclose(fic_out);
        free(phrase);
        free(meilleur_chemin);
        free(meilleur_chemin_lisse);
        free(p_test);
        printf("\n");
        return 0;
      }
      taille ++;
      phrase[taille-1] = mot;
      progression += strlen(mot)+1;
    } while (strcmp(mot,"</s>") != 0);
    /* Etiqueter la phrase */
    meilleur_chemin = viterbi(pi,dico_ht,bigram,phrase,taille,nb_cat);
    for (i=0 ; i<taille ; i++) {
      // Etiquetage avec T(i,j) normale
      best = meilleur_chemin[i];
      categorie = tab_cat[best];
      fprintf(fic_out,"%s %s\n",phrase[i],categorie);
    }
    /* Affichage de la progression */
    if (total > 0) {
      while (palier+2 < (int)(progression*100/total)) {
        palier += 2;
        printf("|");
        fflush(stdout);
      }
    }
    /* Liberation de la memoire */
    for (i=0 ; i<taille ; i++) {
      free(phrase[i]);
    }
    free(meilleur_chemin);
  }
  printf("\n");
  return 0;
}
