#include "viterbi.h"

int *viterbi (double *pi, sdg_ht *ht_emis, double **t_trans, char **phrase, int long_phrase, int nb_cat) {
  int t,i;
  double **delta;
  int **psi;
  int *meilleur_chemin;

  /* Init de delta */
  delta = malloc(long_phrase*sizeof(double*));
  for (t=0 ; t<long_phrase ; t++) {
    delta[t] = malloc(nb_cat*sizeof(double));
  }
  initialise_delta(pi,ht_emis,phrase,delta,nb_cat);
  /* Init de psi */
  psi = malloc(long_phrase*sizeof(int*));
  for (t=0 ; t<long_phrase ; t++) {
    psi[t] = malloc(nb_cat*sizeof(int));
  }
  /* Calcule de delta */
  calcule_delta(delta,ht_emis,t_trans,psi,phrase,long_phrase,nb_cat);
  /* Calcule du meilleur chemin */
  meilleur_chemin = malloc(long_phrase*sizeof(int));
  determiner_meilleur_chemin(delta,psi,meilleur_chemin,long_phrase,nb_cat);
  /* Liberation mem */
  for (i=0 ; i<long_phrase ; i++) {
    free(delta[i]);
  }
  free(delta);
  for (i=0 ; i<long_phrase ; i++) {
    free(psi[i]);
  }
  free(psi);
  
  return meilleur_chemin;
}

void initialise_delta (double *pi, sdg_ht *ht_emis, char **phrase, double **delta,  int nb_cat) {
  int i;

  /**
   * delta[0][i] = p(m,c[i])
   *             = p(m|c[i]) x p(c[i])
   *   o -p(m|c[i]) = E(m,c[i]) = proba_mot_cat(ht_emis, phrase[0], i)
   *      -p(c[i]) = 1/nb_cat
   *   avec -m     : le 1er mot de la phrase (m=phrase[0])
   *        -c[i]  : la catgorie d'indice i
   *        -nb_cat: le nombre de catgories
   * Rem: Pour p(c[i]) on aurait pu faire l'hypothse qu'une phrase ne peut commencer que par '<s>'
   *      Dans ce cas p(c[i]) = {0 pour tout i!=44 (ZTRM)
   *                            {1 pour i=44
   *      Le resultat aurait t le mme pour l'tiquetage
   */
  for (i=0 ; i<nb_cat ; i++) {
    delta[0][i] = proba_mot_cat(ht_emis, phrase[0], i) * pi[i];
  }
  return;
}

void calcule_delta (double **delta, sdg_ht *ht_emis, double **t_trans, int **psi, char **phrase, int long_phrase, int nb_cat) {
  int i,j,t;
  double max,proba;

  for (t=1 ; t<long_phrase ; t++) {
    for (j=0 ; j<nb_cat ; j++) {
      max = -1;
      for (i=0 ; i<nb_cat ; i++) {
        proba = (double)delta[t-1][i];
        proba *= (double)(proba_cat_cat(t_trans,i,j));
        proba *= (double)(proba_mot_cat(ht_emis,phrase[t],j));
        if (max < proba) {
          max = proba;
          psi[t][j] = i;
        }
      }
      delta[t][j] = max;
    }
  }
  return;
}

void determiner_meilleur_chemin (double **delta, int **psi, int *meilleur_chemin,  int long_phrase, int nb_cat) {
  int t,i,best;
  double max = -1;

  /* Recherche du dernier tat le plus probable */
  for (i=0 ; i<nb_cat ; i++) {
    if (delta[long_phrase-1][i] > max) {
      max = delta[long_phrase-1][i];
      best = i;
    }
  }
  meilleur_chemin[long_phrase-1] = best;
  /* Commence par le dernier mot et remonte jusqu'au 1er */
  for (t=long_phrase-1 ; t>0 ; t--) {
    meilleur_chemin[t-1] = psi[t][meilleur_chemin[t]];
  }
  return;
}
